# Copyright (c) Facebook, Inc. and its affiliates. (https://www.facebook.com)
# Copyright (c) 2022 Mapillary, Inc. (https://www.mapillary.com)
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# -*- coding: utf-8 -*-

"""
mapillary.utils.logger
~~~~~~~~~~~~~~~~~~~~~~

This module implements the logger for mapillary, which is a wrapper of the logger package and
the default configuration for each of the loggers per file.

"""

import logging
import sys
import os


class Logger:

    format_string: str = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    level: int = logging.INFO

    @staticmethod
    def setup_logger(name: str, level: int = logging.INFO) -> logging.Logger:
        """
        Function setup as many loggers as you want. To be used at the top of the file.

        Usage::

            >>> Logger.setup_logger(name='mapillary.xxxx.yyyy', level=logging.INFO)
            logger.Logger

        :param name: The name of the logger
        :type name: str

        :param level: The level of the logger
        :type level: int

        :return: The logger object
        :rtype: logging.Logger
        """

        # Basic logger setup
        logger: Logger = logging.getLogger(name)

        # stdout logger setup
        stream_handler: logging.StreamHandler = logging.StreamHandler(sys.stdout)
        logger.addHandler(stream_handler)

        try:
            logger.setLevel(level)
        except ValueError:
            logger.setLevel(logging.INFO)
            logger.warning("LOG_LEVEL: Invalid variable - defaulting to: INFO")

        # File logger setup
        formatter: logging.Formatter = logging.Formatter(Logger.format_string)
        handler: logging.FileHandler = logging.FileHandler(
            Logger.get_os_log_path(name.split(".")[1] + ".log")
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)

        return logger

    @staticmethod
    def get_os_log_path(log_file: str) -> str:
        """
        Get the path of the log file based on the OS

        :param log_file: The name of the log file
        :type log_file: str

        :return: The path where the logs will be stored
        :rtype: str
        """

        log_path: str = os.path.join(os.path.dirname(os.path.abspath(__file__)), "logs")

        if not os.path.exists(log_path):
            os.makedirs(log_path)

        return os.path.join(log_path, log_file)



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controllers.image
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the image filtering and analysis business logic functionalities of the
Mapillary Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Library imports
import json

import mercantile
import shapely
from geojson import Polygon
from typing import Union

# # Configs
from zensvi.download.mapillary.config.api.entities import Entities
from zensvi.download.mapillary.config.api.vector_tiles import VectorTiles
from zensvi.download.mapillary.models.api.entities import EntityAdapter
from zensvi.download.mapillary.models.api.general import GeneralAdapter

# # Adapters
from zensvi.download.mapillary.models.api.vector_tiles import VectorTilesAdapter

# # Client
from zensvi.download.mapillary.models.client import Client

# # Exception Handling
from zensvi.download.mapillary.models.exceptions import InvalidImageKeyError

# # Class Representation
from zensvi.download.mapillary.models.geojson import GeoJSON, Coordinates

# # Utilities
from zensvi.download.mapillary.utils.filter import pipeline
from zensvi.download.mapillary.utils.format import (
    feature_to_geojson,
    merged_features_list_to_geojson,
    geojson_to_polygon,
    coord_or_list_to_dict,
)
from zensvi.download.mapillary.utils.verify import (
    image_check,
    image_bbox_check,
    sequence_bbox_check,
    resolution_check,
    valid_id,
)
from requests import HTTPError
from turfpy.measurement import bbox
from vt2geojson.tools import vt_bytes_to_geojson


def get_image_close_to_controller(
    longitude: float,
    latitude: float,
    kwargs: dict,
) -> GeoJSON:
    """
    Extracting the GeoJSON for the image data near the [longitude, latitude] coordinates

    :param kwargs: The kwargs for the filter
    :type kwargs: dict

    :param longitude: The longitude
    :type longitude: float

    :param latitude: The latitude
    :type latitude: float

    :param kwargs.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type kwargs.zoom: int

    :param kwargs.min_captured_at: The minimum date to filter till
    :type kwargs.min_captured_at: str

    :param kwargs.max_captured_at: The maximum date to filter upto
    :type kwargs.max_captured_at: str

    :param kwargs.image_type: Either 'pano', 'flat' or 'all'
    :type kwargs.image_type: str

    :param kwargs.organization_id: The organization to retrieve the data for
    :type kwargs.organization_id: str

    :param kwargs.radius: The radius that the geometry points will lie in
    :type kwargs.radius: float

    :return: GeoJSON
    :rtype: dict
    """

    # Checking if a non valid key has been passed to the function If that is the case, throw an
    # exception
    image_check(kwargs=kwargs)

    unfiltered_data = VectorTilesAdapter().fetch_layer(
        layer="image",
        zoom=kwargs["zoom"] if "zoom" in kwargs else 14,
        longitude=longitude,
        latitude=latitude,
    )

    if kwargs == {}:
        return GeoJSON(geojson=unfiltered_data)

    # Filtering for the attributes obtained above
    if (
        len(unfiltered_data["features"]) != 0
        and unfiltered_data["features"][0]["properties"] != {}
    ):
        return GeoJSON(
            geojson=json.loads(
                merged_features_list_to_geojson(
                    pipeline(
                        data=unfiltered_data,
                        components=[
                            # Filter using kwargs.min_captured_at
                            {
                                "filter": "min_captured_at",
                                "min_timestamp": kwargs["min_captured_at"],
                            }
                            if "min_captured_at" in kwargs
                            else {},
                            # Filter using kwargs.max_captured_at
                            {
                                "filter": "max_captured_at",
                                "min_timestamp": kwargs["max_captured_at"],
                            }
                            if "max_captured_at" in kwargs
                            else {},
                            # Filter using kwargs.image_type
                            {"filter": "image_type", "tile": kwargs["image_type"]}
                            if "image_type" in kwargs
                            else {},
                            # Filter using kwargs.organization_id
                            {
                                "filter": "organization_id",
                                "organization_ids": kwargs["organization_id"],
                            }
                            if "organization_id" in kwargs
                            else {},
                            # Filter using kwargs.radius
                            {
                                "filter": "haversine_dist",
                                "radius": kwargs["radius"],
                                "coords": [longitude, latitude],
                            }
                            if "radius" in kwargs
                            else {},
                        ],
                    )
                )
            )
        )


def get_image_looking_at_controller(
    at: Union[dict, Coordinates, list],
    filters: dict,
) -> GeoJSON:
    """
    Checks if the image with coordinates 'at' is looked with the given filters.

    :param filters: Filters to pass the data through
    :type filters: dict

    :param at: The dict of coordinates of the position of the looking at
        coordinates. Format::

            >>> {
            >>>     'lng': 'longitude',
            >>>     'lat': 'latitude'
            >>> }

    :type at: dict

    :param filters.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type filters.zoom: int

    :param filters.min_captured_at: The minimum date to filter till
    :type filters.min_captured_at: str

    :param filters.max_captured_at: The maximum date to filter upto
    :type filters.max_captured_at: str

    :param filters.radius: The radius that the geometry points will lie in
    :type filters.radius: float

    :param filters.image_type: Either 'pano', 'flat' or 'all'
    :type filters.image_type: str

    :param filters.organization_id: The organization to retrieve the data for
    :type filters.organization_id: str

    :return: GeoJSON
    :rtype: dict
    """

    # Converting 'at' of type Coordinates|List to dict
    at: dict = coord_or_list_to_dict(data=at)

    # Checking if a non valid key
    # has been passed to  the function
    # If that is the case, throw an exception
    image_check(kwargs=filters)

    at_image_data = GeneralAdapter().fetch_image_tiles(
        zoom=filters["zoom"] if "zoom" in filters else 14,
        longitude=at["lng"],
        latitude=at["lat"],
        layer="image",
    )

    if not at_image_data["features"]:
        return GeoJSON(geojson=at_image_data)

    # Filters are to be applied to the data retrieved from the database in the following logic
    # # 1. Filter by the filters provided by the filters parameter
    # # 2. Secondly, from the fetched tile, trim out data that falls outside the given radius, if
    # no radius is specified, assume 20m to be safe
    # # 2. Then, from the remaining feature points, extract only those that are qualified by the
    # "hits_by_look_at" function

    # Filter the unfiltered results by the given filters
    return GeoJSON(
        geojson=json.loads(
            merged_features_list_to_geojson(
                pipeline(
                    data=at_image_data,
                    components=[
                        # Filter by `max_captured_at`
                        {
                            "filter": "max_captured_at",
                            "max_timestamp": filters.get("max_captured_at"),
                        }
                        if "max_captured_at" in filters
                        else {},
                        # Filter by `min_captured_at`
                        {
                            "filter": "min_captured_at",
                            "min_timestamp": filters.get("min_captured_at"),
                        }
                        if "min_captured_at" in filters
                        else {},
                        # Filter by `image_type`
                        {"filter": "image_type", "type": filters.get("image_type")}
                        if "image_type" in filters and filters["image_type"] != "all"
                        else {},
                        # Filter by `organization_id`
                        {
                            "filter": "organization_id",
                            "organization_ids": filters.get("organization_id"),
                        }
                        if "organization_id" in filters
                        else {},
                        # Filter using kwargs.radius
                        {
                            "filter": "haversine_dist",
                            "radius": filters.get("radius"),
                            "coords": [at["lng"], at["lat"]],
                        }
                        if "radius" in filters
                        else {},
                        # Filter by `hits_by_look_at`
                        {"filter": "hits_by_look_at", "at": at},
                    ],
                )
            )
        )
    )


def is_image_being_looked_at_controller(
    at: Union[dict, Coordinates, list],
    filters: dict,
) -> bool:
    """
    Checks if the image with coordinates 'at' is looked with the given filters.

    :param at: The dict of coordinates of the position of the looking at coordinates.

        Format::

            >>> at_dict = {
            ...     'lng': 'longitude',
            ...     'lat': 'latitude'
            ... }
            >>> at_list = [12.954940544167, 48.0537894275]
            >>> from mapillary.models.geojson import Coordinates
            >>> at_coord: Coordinates = Coordinates(lng=12.954940544167, lat=48.0537894275)

    :type at: Union[dict, mapillary.models.geojson.Coordinates, list]

    :param filters.zoom: The zoom level of the tiles to obtain, defaults to 14
    :type filter.zoom: int

    :param filters.min_captured_at: The minimum date to filter till
    :type filters.min_captured_at: str

    :param filters.max_captured_at: The maximum date to filter upto
    :type filters.max_captured_at: str

    :param filters.radius: The radius that the geometry points will lie in
    :type filters.radius: float

    :param filters.image_type: Either 'pano', 'flat' or 'all'
    :type filters.image_type: str

    :param filters.organization_id: The organization to retrieve the data for
    :type filters.organization_id: str

    :return: True if the image is looked at by the given looker and at coordinates, False otherwise
    :rtype: bool
    """

    result: dict = get_image_looking_at_controller(at=at, filters=filters).to_dict()

    # If the result is empty, the image is not looked at, hence return False
    return len(result["features"]) != 0


def get_image_thumbnail_controller(image_id: str, resolution: int) -> str:
    """
    This controller holds the business logic for retrieving
    an image thumbnail with a specific resolution (256, 1024, or 2048)
    using an image ID/key

    :param image_id: Image key as the argument
    :type image_id: str

    :param resolution: Option for the thumbnail size, with available resolutions:
        256, 1024, and 2048
    :type resolution: int

    :return: A URL for the thumbnail
    :rtype: str
    """

    # check if the entered resolution is one of the supported image sizes
    resolution_check(resolution)

    try:
        res = Client().get(Entities.get_image(image_id, [f"thumb_{resolution}_url"]))
    except HTTPError:
        # If given ID is an invalid image ID, let the user know
        raise InvalidImageKeyError(image_id)

    return json.loads(res.content.decode("utf-8"))[f"thumb_{resolution}_url"]


def get_images_in_bbox_controller(
    bounding_box: dict, layer: str, zoom: int, filters: dict
) -> str:
    """
    For getting a complete list of images that lie within a bounding box,
    that can be filtered via the filters argument

    :param bounding_box: A bounding box representation
        Example::

            >>> {
            ...     'west': 'BOUNDARY_FROM_WEST',
            ...     'south': 'BOUNDARY_FROM_SOUTH',
            ...     'east': 'BOUNDARY_FROM_EAST',
            ...     'north': 'BOUNDARY_FROM_NORTH'
            ... }

    :type bounding_box: dict

    :param zoom: The zoom level
    :param zoom: int

    :param layer: Either 'image', 'sequence', 'overview'
    :type layer: str

    :param filters: Filters to pass the data through
    :type filters: dict

    :param filters.max_captured_at: The max date that can be filtered upto
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date that can be filtered from
    :type filters.min_captured_at: str

    :param filters.image_type: Either 'pano', 'flat' or 'all'
    :type filters.image_type: str

    :param filters.compass_angle:
    :type filters.compass_angle: float

    :param filters.organization_id:
    :type filters.organization_id: int

    :param filters.sequence_id:
    :type filters.sequence_id: str

    :raises InvalidKwargError: Raised when a function is called with the invalid keyword argument(s)
        that do not belong to the requested API end call

    :return: GeoJSON
    :rtype: str

    Reference,

    - https://www.mapillary.com/developer/api-documentation/#coverage-tiles
    """

    # Check if the given filters are valid ones
    filters["zoom"] = filters.get("zoom", zoom)
    filters = (
        image_bbox_check(filters) if layer == "image" else sequence_bbox_check(filters)
    )

    # Instantiate the Client
    client = Client()

    # filtered images or sequence data will be appended to this list
    filtered_results = []

    # A list of tiles that are either confined within or intersect with the bbox
    tiles = list(
        mercantile.tiles(
            west=bounding_box["west"],
            south=bounding_box["south"],
            east=bounding_box["east"],
            north=bounding_box["north"],
            zooms=zoom,
        )
    )

    for tile in tiles:
        url = (
            VectorTiles.get_image_layer(x=tile.x, y=tile.y, z=tile.z)
            if layer == "image"
            else VectorTiles.get_sequence_layer(x=tile.x, y=tile.y, z=tile.z)
        )

        # Get the response from the API
        res = client.get(url)

        # Get the GeoJSON response by decoding the byte tile
        geojson = vt_bytes_to_geojson(
            b_content=res.content, layer=layer, z=tile.z, x=tile.x, y=tile.y
        )

        # Filter the unfiltered results by the given filters
        filtered_results.extend(
            pipeline(
                data=geojson,
                components=[
                    {"filter": "features_in_bounding_box", "bbox": bounding_box}
                    if layer == "image"
                    else {},
                    {
                        "filter": "max_captured_at",
                        "max_timestamp": filters.get("max_captured_at"),
                    }
                    if filters["max_captured_at"] is not None
                    else {},
                    {
                        "filter": "min_captured_at",
                        "min_timestamp": filters.get("min_captured_at"),
                    }
                    if filters["min_captured_at"] is not None
                    else {},
                    {"filter": "image_type", "type": filters.get("image_type")}
                    if filters["image_type"] is not None
                    or filters["image_type"] != "all"
                    else {},
                    {
                        "filter": "organization_id",
                        "organization_ids": filters.get("organization_id"),
                    }
                    if filters["organization_id"] is not None
                    else {},
                    {"filter": "sequence_id", "ids": filters.get("sequence_id")}
                    if layer == "image" and filters["sequence_id"] is not None
                    else {},
                    {"filter": "compass_angle", "angles": filters.get("compass_angle")}
                    if layer == "image" and filters["compass_angle"] is not None
                    else {},
                ],
            )
        )

    return merged_features_list_to_geojson(filtered_results)


def get_image_from_key_controller(key: int, fields: list) -> str:
    """
    A controller for getting properties of a certain image given the image key and
    the list of fields/properties to be returned

    :param key: The image key
    :type key: int

    :param fields: The list of fields to be returned
    :type fields: list

    :return: The requested image properties in GeoJSON format
    :rtype: str
    """

    valid_id(identity=key, image=True)

    # ? 'merged_features_list_to_geojson' takes list, 'feature_to_geojson' returns dict
    return merged_features_list_to_geojson(
        features_list=feature_to_geojson(
            json_data=EntityAdapter().fetch_image(image_id=key, fields=fields)
        )
    )


def geojson_features_controller(
    geojson: dict, is_image: bool = True, filters: dict = None
) -> GeoJSON:
    """
    For extracting images that lie within a GeoJSON and merges the results of the found
    GeoJSON(s) into a single object - by merging all the features into one feature list.

    :param geojson: The geojson to act as the query extent
    :type geojson: dict

    :param is_image: Is the feature extraction for images? True for images, False for map features
        Defaults to True
    :type is_image: bool

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.zoom: The zoom level to obtain vector tiles for, defaults to 14
    :type filters.zoom: int

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :param filters.layer: The specified image layer, either 'overview', 'sequence', 'image'
        if is_image is True, defaults to 'image'
    :type filters.layer: str

    :param filters.feature_type: The specified map features, either 'point' or 'traffic_signs'
        if is_image is False, defaults to 'point'
    :type filters.feature_type: str

    :raises InvalidKwargError: Raised when a function is called with the invalid keyword argument(s)
        that do not belong to the requested API end call

    :return: A feature collection as a GeoJSON
    :rtype: dict
    """

    # Filter checking
    image_bbox_check(filters)

    # Extracting polygon from geojson, converting to dict
    polygon = geojson_to_polygon(geojson).to_dict()

    # Generating a coordinates list to extract from polygon
    coordinates_list = []

    # Going through each feature
    for feature in polygon["features"]:

        # Going through the coordinate's nested list
        for coordinates in feature["geometry"]["coordinates"][0]:
            # Appending a tuple of coordinates
            coordinates_list.append((coordinates[0], coordinates[1]))

    # Sending coordinates_list a input to form a Polygon
    polygon = Polygon(coordinates_list)

    # Getting the boundary parameters from polygon
    boundary = shapely.geometry.shape(polygon)

    if is_image:
        # Get a GeoJSON with features from tiles originating from coordinates
        # at specified zoom level
        layers: dict = (
            VectorTilesAdapter()
            .fetch_layers(
                # Sending coordinates for all the points within input geojson
                coordinates=bbox(polygon),
                # Fetching image layers for the geojson
                layer=filters["layer"] if "layer" in filters else "image",
                # Specifying zoom level, defaults to zoom if zoom not specified
                zoom=filters["zoom"] if "zoom" in filters else 14,
            )
            .to_dict()
        )
    else:
        # Get all the map features within the boundary box for the polygon
        layers: dict = (
            VectorTilesAdapter()
            .fetch_map_features(
                # Sending coordinates for all the points within input geojson
                coordinates=bbox(polygon),
                # Fetching image layers for the geojson
                feature_type=filters["feature_type"]
                if "feature_type" in filters
                else "point",
                # Specifying zoom level, defaults to zoom if zoom not specified
                zoom=filters["zoom"] if "zoom" in filters else 14,
            )
            .to_dict()
        )

    # Return as GeoJSON output
    return GeoJSON(
        # Load the geojson to convert to GeoJSON object
        geojson=json.loads(
            # Convert feature list to GeoJSON
            merged_features_list_to_geojson(
                # Execute pipeline for filters
                pipeline(
                    # Sending layers as input
                    data=layers,
                    # Specifying components for the filter
                    components=[
                        {"filter": "in_shape", "boundary": boundary},
                        # Filter using kwargs.min_captured_at
                        {
                            "filter": "min_captured_at",
                            "min_timestamp": filters["min_captured_at"],
                        }
                        if "min_captured_at" in filters
                        else {},
                        # Filter using filters.max_captured_at
                        {
                            "filter": "max_captured_at",
                            "min_timestamp": filters["max_captured_at"],
                        }
                        if "max_captured_at" in filters
                        else {},
                        # Filter using filters.image_type
                        {"filter": "image_type", "tile": filters["image_type"]}
                        if "image_type" in filters
                        else {},
                        # Filter using filters.organization_id
                        {
                            "filter": "organization_id",
                            "organization_ids": filters["org_id"],
                        }
                        if "organization_id" in filters
                        else {},
                        # Filter using filters.sequence_id
                        {"filter": "sequence_id", "ids": filters.get("sequence_id")}
                        if "sequence_id" in filters
                        else {},
                        # Filter using filters.compass_angle
                        {
                            "filter": "compass_angle",
                            "angles": filters.get("compass_angle"),
                        }
                        if "compass_angle" in filters
                        else {},
                    ],
                )
            )
        )
    )


def shape_features_controller(
    shape, is_image: bool = True, filters: dict = None
) -> GeoJSON:
    """
    For extracting images that lie within a shape, merging the results of the found features
    into a single object - by merging all the features into one list in a feature collection.

    The shape format is as follows::

        >>> {
        ...     "type": "FeatureCollection",
        ...     "features": [
        ...         {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                        [
        ...                              7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

    :param shape: A shape that describes features, formatted as a geojson
    :type shape: dict

    :param is_image: Is the feature extraction for images? True for images, False for map features
        Defaults to True
    :type is_image: bool

    :param filters: Different filters that may be applied to the output, defaults to {}
    :type filters: dict (kwargs)

    :param filters.max_captured_at: The max date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.max_captured_at: str

    :param filters.min_captured_at: The min date. Format from 'YYYY', to 'YYYY-MM-DDTHH:MM:SS'
    :type filters.min_captured_at: str

    :param filters.image_type: The tile image_type to be obtained, either as 'flat', 'pano'
        (panoramic), or 'all'. See https://www.mapillary.com/developer/api-documentation/ under
        'image_type Tiles' for more information
    :type filters.image_type: str

    :param filters.compass_angle: The compass angle of the image
    :type filters.compass_angle: int

    :param filters.sequence_id: ID of the sequence this image belongs to
    :type filters.sequence_id: str

    :param filters.organization_id: ID of the organization this image belongs to. It can be absent
    :type filters.organization_id: str

    :param filters.layer: The specified image layer, either 'overview', 'sequence', 'image'
        if is_image is True, defaults to 'image'
    :type filters.layer: str

    :param filters.feature_type: The specified map features, either 'point' or 'traffic_signs'
        if is_image is False, defaults to 'point'
    :type filters.feature_type: str

    :raises InvalidKwargError: Raised when a function is called with the invalid keyword argument(s)
        that do not belong to the requested API end call

    :return: A feature collection as a GeoJSON
    :rtype: dict
    """

    image_bbox_check(filters)

    # Generating a coordinates list to extract from polygon
    coordinates_list = []

    # Going through each feature
    for feature in shape["features"]:

        # Going through the coordinate's nested list
        for coordinates in feature["geometry"]["coordinates"][0]:
            # Appending a tuple of coordinates
            coordinates_list.append((coordinates[0], coordinates[1]))

    # Sending coordinates_list a input to form a Polygon
    polygon = Polygon(coordinates_list)

    # Getting the boundary parameters from polygon
    boundary = shapely.geometry.shape(polygon)

    if is_image:
        # Get all the map features within the boundary box for the polygon
        output: dict = (
            VectorTilesAdapter()
            .fetch_layers(
                # Sending coordinates for all the points within input geojson
                coordinates=bbox(polygon),
                # Fetching image layers for the geojson
                layer=filters["layer"] if "layer" in filters else "image",
                # Specifying zoom level, defaults to zoom if zoom not specified
                zoom=filters["zoom"] if "zoom" in filters else 14,
            )
            .to_dict()
        )
    else:
        # Get all the map features within the boundary box for the polygon
        output: dict = (
            VectorTilesAdapter()
            .fetch_map_features(
                # Sending coordinates for all the points within input geojson
                coordinates=bbox(polygon),
                # Fetching image layers for the geojson
                feature_type=filters["feature_type"]
                if "feature_type" in filters
                else "point",
                # Specifying zoom level, defaults to zoom if zoom not specified
                zoom=filters["zoom"] if "zoom" in filters else 14,
            )
            .to_dict()
        )

    # Return as GeoJSON output
    return GeoJSON(
        # Load the geojson to convert to GeoJSON object
        geojson=json.loads(
            # Convert feature list to GeoJSON
            merged_features_list_to_geojson(
                # Execute pipeline for filters
                pipeline(
                    # Sending layers as input
                    data=output,
                    # Specifying components for the filter
                    components=[
                        # Get only features within the given boundary
                        {"filter": "in_shape", "boundary": boundary},
                        # Filter using kwargs.min_captured_at
                        {
                            "filter": "min_captured_at",
                            "min_timestamp": filters["min_captured_at"],
                        }
                        if "min_captured_at" in filters
                        else {},
                        # Filter using filters.max_captured_at
                        {
                            "filter": "max_captured_at",
                            "min_timestamp": filters["max_captured_at"],
                        }
                        if "max_captured_at" in filters
                        else {},
                        # Filter using filters.image_type
                        {"filter": "image_type", "tile": filters["image_type"]}
                        if "image_type" in filters
                        else {},
                        # Filter using filters.organization_id
                        {
                            "filter": "organization_id",
                            "organization_ids": filters["org_id"],
                        }
                        if "organization_id" in filters
                        else {},
                        # Filter using filters.sequence_id
                        {"filter": "sequence_id", "ids": filters.get("sequence_id")}
                        if "sequence_id" in filters
                        else {},
                        # Filter using filters.compass_angle
                        {
                            "filter": "compass_angle",
                            "angles": filters.get("compass_angle"),
                        }
                        if "compass_angle" in filters
                        else {},
                    ],
                )
            )
        )
    )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controller.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This package contains all the business logic implementations of interfaces used within the
Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import detection  # noqa: F401
from . import feature  # noqa: F401
from . import image  # noqa: F401
from . import save  # noqa: F401



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.api.entities
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the Adapter design for the Entities API of Mapillary API v4.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import mercantile
import typing
from vt2geojson.tools import vt_bytes_to_geojson

# Local imports

# # Models
from zensvi.download.mapillary.models.client import Client

# # Exception Handling
from zensvi.download.mapillary.models.exceptions import InvalidOptionError, LiteralEnforcementException

# # Config
from zensvi.download.mapillary.config.api.general import General

# Library imports
from requests import HTTPError


class GeneralAdapter(object):
    """
    General adaptor for using the API calls defined for the general module
    (mapillary.config.api.general)

    The GeneralAdaptor provides functions for getting preprocessed data from the API, through the
    API calls mentioned in the previously mentioned config.

    It performs parsing, property handling, easier logic for extracing information, focusing on
    adding a layer of abstraction by removing details of using `mercantile`, `ast`, et al, to
    focus on the inputs and outputs of functions

    Usage::

        >>> import mapillary
    """

    # layer types
    __LAYER_TYPES: typing.List[str] = [
        "overview",
        "sequence",
        "image",
        "map_feature",
        "traffic_sign",
    ]

    def __init__(self, *args: object) -> None:
        """Initializing GeneralAdaptor constuctor"""
        super().__init__(*args)

        # client object to deal with session and requests
        self.client = Client()

        # Setting the max zoom value
        self.__min_zoom = 0

        # Setting the min zoom value
        self.__max_zoom = 14

    def fetch_image_tiles(
        self,
        zoom: int,
        longitude: float,
        latitude: float,
        layer: str = "image",
    ) -> dict:
        """
        Get the tiles for a given image.

        :param zoom: Zoom level of the image.
        :type zoom: int

        :param longitude: Longitude of the image
        :type longitude: float

        :param latitude: Latitude of the image
        :type latitude: float

        :return: A dictionary containing the tiles for the image.
        :rtype: dict
        """

        # Perform validation checks
        self.__validation_checks(
            zoom=zoom,
            longitude=longitude,
            latitude=latitude,
            layer=layer,
        )

        # Get the tiles for the image
        return self.__preprocess_layer(
            longitude=longitude,
            latitude=latitude,
            zoom=zoom,
            feature_type="images",
            layer=layer,
        )

    def fetch_computed_image_tiles(
        self,
        zoom: int,
        longitude: float,
        latitude: float,
        layer: str = "image",
    ) -> dict:
        """
        Get the image type for a given image.

        :param zoom: The zoom to get the image type for.
        :type zoom: int

        :param longitude: The longitude of the image.
        :type longitude: float

        :param latitude: The latitude of the image.
        :type latitude: float

        :return: A dictionary containing the image type for the image.
        :rtype: dict
        """
        # Perform validation checks
        self.__validation_checks(
            zoom=zoom,
            longitude=longitude,
            latitude=latitude,
            layer=layer,
        )

        # Get the image type for the image
        return self.__preprocess_layer(
            longitude=longitude,
            latitude=latitude,
            zoom=zoom,
            feature_type="images",
            layer=layer,
            is_computed=True,
        )

    def fetch_map_features_tiles(
        self,
        zoom: int,
        longitude: float,
        latitude: float,
        layer: str = "image",
    ):
        """
        Get the map features for a given coordinate set

        :param zoom: The zoom value to get the map features for
        :type zoom: int

        :param longitude: The longitude of the image
        :type longitude: float

        :param latitude: The latitude of the image
        :type latitude: float

        :return: A dictionary containing the map features for the image.
        :rtype: dict
        """
        # Perform validation checks
        self.__validation_checks(
            zoom=zoom,
            longitude=longitude,
            latitude=latitude,
            layer=layer,
        )

        # Get the map features for the image
        return self.__preprocess_layer(
            longitude=longitude,
            latitude=latitude,
            zoom=zoom,
            feature_type="map_features__points",
            layer=layer,
        )

    def fetch_map_features_traffic_tiles(
        self, zoom: int, longitude: float, latitude: float, layer: str
    ):
        """
        Get the map feature traffic for a given coordinate set

        :param zoom: The zoom value to get the map features for
        :type zoom: int

        :param longitude: The longitude of the image
        :type longitude: float

        :param latitude: The latitude of the image
        :type latitude: float

        :return: A dictionary containing the map features for the image.
        :rtype: dict
        """
        # Perform validation checks
        self.__validation_checks(
            zoom=zoom,
            longitude=longitude,
            latitude=latitude,
            layer=layer,
        )

        # Get the map features for the image
        return self.__preprocess_layer(
            longitude=longitude,
            latitude=latitude,
            zoom=zoom,
            feature_type="map_features__traffic_signs",
            layer=layer,
        )

    def __preprocess_layer(
        self,
        longitude: float,
        latitude: float,
        zoom: int = 14,
        feature_type: str = "images",
        layer: str = "image",
        is_computed: bool = False,
    ) -> any:
        try:
            tile: mercantile.Tile = mercantile.tile(
                lng=longitude, lat=latitude, zoom=zoom
            )

            return vt_bytes_to_geojson(
                # Parameters appropriately
                b_content=self.client.get(
                    self.__preprocess_api_string(
                        # Turn coordinates into a tile
                        tile=mercantile.tile(lng=longitude, lat=latitude, zoom=zoom),
                        # the layer to retrieve from
                        layer=feature_type,
                        # is the layer computed
                        is_computed=is_computed,
                    )
                ).content,
                x=tile.x,
                y=tile.y,
                z=tile.z,
                layer=layer,
            )
        except HTTPError as e:
            raise HTTPError(e)

    def __preprocess_api_string(
        self,
        tile: mercantile.Tile,
        layer: str = "images",
        is_computed: bool = False,
    ) -> str:
        """
        Preprocess the API string for the given tile.

        :param tile: The tile to preprocess the API string for.
        :type tile: mercantile.Tile

        :param layer: The feature type to preprocess the API string for.
        :type layer: str

        :param is_computed: Whether the API string is for a computed image or not.
        :type is_computed: bool

        :return: The preprocessed API string.
        :rtype: str
        """

        if layer == "images":
            # Get the API string for the tile
            if is_computed:
                # If the API string is for a computed image
                return General.get_computed_image_type_tiles(
                    # The zoom value
                    z=tile.z,
                    # The longitude value
                    x=tile.x,
                    # The latitude value
                    y=tile.y,
                )
            else:
                # If the API string is for an image
                return General.get_image_type_tiles(
                    # The zoom value
                    z=tile.z,
                    # The longitude value
                    x=tile.x,
                    # The latitude value
                    y=tile.y,
                )

        elif layer == "map_features__points":
            # Get the API string for the tile
            return General.get_map_features_points_tiles(
                # The zoom value
                z=tile.z,
                # The longitude value
                x=tile.x,
                # The latitude value
                y=tile.y,
            )

        elif layer == "map_features__traffic_signs":
            # Get the API string for the tile
            return General.get_map_features_traffic_signs_tiles(
                # The zoom value
                z=tile.z,
                # The longitude value
                x=tile.x,
                # The latitude value
                y=tile.y,
            )

        else:
            raise InvalidOptionError(
                param="layer",
                value=layer,
                options=[
                    "images",
                    "map_features__points",
                    "map_features__traffic_signs",
                ],
            )

    @staticmethod
    def __check_parameters(
        longitude: float,
        latitude: float,
    ):
        """
        Range checking for the parameters of longitude, latitude, layer, zoom

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :return: A GeoJSON for that specific layer and the specified zoom level
        :rtype: dict
        """

        # Lng, Lat ranges, https://docs.mapbox.com/help/glossary/lat-lon/

        # If lng not in the range [-180, 180], inclusive
        if longitude <= -180 or longitude >= 180:
            # Raise exception
            raise InvalidOptionError(
                param="longitude", value=longitude, options=[-180, 180]
            )

        # If lat not in the range [-90, 90], inclusive
        if latitude <= -90 or latitude >= 90:
            # Raise exception
            raise InvalidOptionError(
                param="latitude", value=latitude, options=[-180, 180]
            )

    def __zoom_range_check(self, layer: str, zoom: int):

        """
        Checks for the correct zoom values for te specified layer

        :param layer: Either 'overview', 'sequence', 'image', or 'map'
        :type layer: str

        :param zoom: The zoom levels,
        :type zoom: int

        :raises InvalidOptionError: Invalid option passed

        :return: A GeoJSON for the return object
        :rtype: dict
        """

        # If zoom is not in the valid range of values
        if zoom < self.__min_zoom or zoom > self.__max_zoom:
            # Raise an exception for the invalid values passed
            raise InvalidOptionError(
                # Parameters accordingly
                param="zoom",
                value=zoom,
                options=[_ for _ in range(self.__min_zoom, self.__max_zoom + 1)],
            )

        # If layer was specified to be 'overview'
        if layer == "overview":

            # If zoom is not in the given range of values
            if zoom not in [_ for _ in range(self.__min_zoom, 5 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(self.__min_zoom, 5 + 1)],
                )

        # If layer was specified to be 'sequence'
        elif layer == "sequence":

            # IF zoom not in the given range of values
            if zoom not in [_ for _ in range(6, 14 + 1)]:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(
                    # Parameters accordingly
                    param="zoom",
                    value=zoom,
                    options=[_ for _ in range(6, 14 + 1)],
                )

        # If layer was specified to be 'image' or 'map'
        elif layer == "image" or layer == "map_feature" or layer == "traffic_sign":

            # If zoom was not 14
            if zoom != 14:
                # Raise an exception for the invalid zoom value
                raise InvalidOptionError(param="zoom", value=zoom, options=[14])

        # Else, an invalid layer string was passed
        else:

            # Raise a InvalidOptionError
            raise InvalidOptionError(
                # Parameters accordingly
                param="layer",
                value=layer,
                options=["overview", "sequence", "image"],
            )

    def __validation_checks(
        self,
        longitude: float,
        latitude: float,
        zoom: int,
        layer: str,
        layer_options: typing.List[str] = [],
    ) -> None:
        """
        Validation checks for the parameters of longitude, latitude, layer, zoom

        :param longitude: The longitude of the coordinates
        :type longitude: float

        :param latitude: The latitude of the coordinates
        :type latitude: float

        :param zoom: The zoom levels,
        :type zoom: int

        :param layer: Either 'overview', 'sequence', 'image', 'map_feature', or 'traffic_sign'
        :type layer: str

        :raises InvalidOptionError: Invalid option passed
        """
        # Check if the correct literals were passed to the function
        LiteralEnforcementException.enforce_literal(
            option_selected=layer,
            options=self.__LAYER_TYPES if layer_options == [] else layer_options,
            param="layer",
        )

        # Check the parameters for the correct range
        self.__check_parameters(longitude, latitude)

        # Check the zoom range for the specified layer
        self.__zoom_range_check(layer, zoom)



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.exceptions
~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the set of Mapillary Exceptions used internally.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import typing


class MapillaryException(Exception):
    """Base class for exceptions in this module"""

    pass


class InvalidBBoxError(MapillaryException):
    """
    Raised when an invalid coordinates for bounding box are entered
    to access Mapillary's API.

    :var message: The error message returned
    :type message: str
    """

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f'InvalidBBoxError: "{self.message}" '

    def __repr__(self):
        return f'InvalidBBoxError: = "{self.message}" '


class InvalidTokenError(MapillaryException):
    """
    Raised when an invalid token is given
    to access Mapillary's API, primarily used in mapillary.set_access_token

    :var message: The error message returned
    :type message: str

    :var error_type: The type of error that occurred
    :type error_type: str

    :var code: The error code returned, most likely 190, "Access token has expired".
        See https://developers.facebook.com/docs/graph-api/using-graph-api/error-handling/
        for more information
    :type code: str

    :var fbtrace_id: A unique ID to track the issue/exception
    :type fbtrace_id: str
    """

    def __init__(self, message: str, error_type: str, code: str, fbtrace_id: str):
        """
        Initializing InvalidTokenError constructor

        :param message: Error message
        :type message: str

        :param error_type: Type of error
        :type error_type: str

        :param code: The error code
        :type code: str

        :param fbtrace_id: the FBTrace_ID
        :type fbtrace_id: str
        """

        self.message = message
        self.error_type = error_type
        self.code = code
        self.fbtrace_id = fbtrace_id

    def __str__(self):
        return f"{self.message}"

    def __repr__(self):
        return (
            "InvalidTokenError: An exception occurred."
            + f'Message: "{self.message}", Type: "{self.error_type}",'
            + f'Code: "{self.code}",'
            + f'fbtrace_id: "{self.fbtrace_id}"'
        )


class AuthError(MapillaryException):
    """
    Raised when a function is called without having the access token set in
    set_access_token to access Mapillary's API, primarily used in mapillary.set_access_token

    :var message: The error message returned
    :type message: str
    """

    def __init__(self, message: str):
        """
        Initializing AuthError constructor

        :param message: Error message
        :type message: str
        """

        self.message = message

    def __str__(self):
        return f'AuthError: An exception occurred, "{self.message}"'

    def __repr__(self):
        return "AuthError: An exception occurred." + f'Message: "{self.message}"'


class InvalidImageResolutionError(MapillaryException):
    """
    Raised when trying to retrieve an image thumbnail with an invalid resolution/size.

    Primarily used with mapillary.image_thumbnail

    :var resolution: Image size entered by the user
    :type resolution: int
    """

    def __init__(self, resolution: int) -> None:
        """
        Initialize InvalidImageResolutionError constructor

        :param resolution: Image resolution
        :type resolution: int
        """

        self._resolution = resolution

    def __str__(self) -> str:
        return f"""An exception occurred, "{self._resolution}" is not a supported image size

Hint: Supported image sizes are: 256, 1024, and 2048
        """

    def __repr__(self) -> str:
        return (
            f'An exception occurred, "{self._resolution}" is not a supported image size'
        )


class InvalidImageKeyError(MapillaryException):
    """
    Raised when trying to retrieve an image thumbnail with an invalid image ID/key.
    Primarily used with mapillary.image_thumbnail

    :var image_id: Image ID/key entered by the user
    :param image_id: int
    """

    def __init__(self, image_id: typing.Union[int, str]) -> None:
        """
        Initializing InvalidImageKeyError constructor

        :param image_id: The image id
        :type image_id: int|str
        """

        self._image_id = image_id

    def __str__(self) -> str:
        return f'An exception occurred, "{self._image_id}" is not a valid image ID/key'

    def __repr__(self) -> str:
        return f'An exception occurred, "{self._image_id}" is not a valid image ID/key'


class InvalidKwargError(MapillaryException):
    """
    Raised when a function is called with the invalid keyword argument(s) that do not belong to the
    requested API end call

    :var func: The function that was called
    :type func: str

    :var key: The key that was passed
    :type key: str

    :var value: The value along with that key
    :type value: str

    :var options: List of possible keys that can be passed
    :type options: list
    """

    def __init__(
        self,
        func: str,
        key: str,
        value: str,
        options: list,
    ):
        """
        Initializing InvalidKwargError constructor

        :param func: The function that was called
        :type func: str

        :param key: The key that was passed
        :type key: str

        :param value: The value along with that key
        :type value: str

        :param options: List of possible keys that can be passed
        :type options: list
        """

        self.func = func
        self.key = key
        self.value = value
        self.options = options

    def __str__(self):
        return (
            f'InvalidKwargError: The invalid kwarg, ["{self.key}": '
            f'{self.value}] was passed to the function, "{self.func}".\n'
            f"A possible list of keys for this function are, "
            f'{", ".join(self.options)}'
        )

    def __repr__(self):
        return (
            f'InvalidKwargError: The invalid kwarg, ["{self.key}": '
            f'{self.value}] was passed to the function, "{self.func}".\n'
            f"A possible list of keys for this function are, "
            f'{", ".join(self.options)}'
        )


class InvalidOptionError(MapillaryException):
    """
    Out of bound zoom error

    :var param: The invalid param passed
    :type param: str

    :var value: The invalid value passed
    :type value: any

    :var options: The possible list of zoom values
    :type options: list
    """

    def __init__(
        self,
        param: str,
        value: any,
        options: list,
    ):
        """
        Initializing InvalidOptionError constructor

        :param param: The invalid param passed
        :type param: str

        :param value: The invalid value passed
        :type value: any

        :param options: The possible list of zoom values
        :type options: list
        """

        self.param = param
        self.value = value
        self.options = options

    def __str__(self):
        return (
            f'InvalidOptionError: Given {self.param} value, "{self.value}" '
            f'while possible {self.param} options, [{", ".join(self.options)}] '
        )

    def __repr__(self):
        return (
            f'InvalidOptionError: Given {self.param} value, "{self.value}" '
            f'while possible {self.param} options, [{", ".join(self.options)}] '
        )


class InvalidFieldError(MapillaryException):
    """
    Raised when an API endpoint is passed invalid field elements

    :var endpoint: The API endpoint that was targeted
    :type endpoint: str

    :var field: The invalid field that was passed
    :type field: list
    """

    def __init__(
        self,
        endpoint: str,
        field: list,
    ):
        """
        Initializing InvalidFieldError constructor

        :param endpoint: The API endpoint that was targeted
        :type endpoint: str

        :param field: The invalid field that was passed
        :type field: list
        """

        self.endpoint = endpoint
        self.field = field

    def __str__(self):
        return (
            f'InvalidFieldError: The invalid field, "{self.field}" was '
            f'passed to the endpoint, "{self.endpoint}"'
        )

    def __repr__(self):
        return (
            f'InvalidFieldError: The invalid field, "{self.field}" was '
            f'passed to the endpoint, "{self.endpoint}"'
        )


class LiteralEnforcementException(MapillaryException):
    """
    Raised when literals passed do not correspond to options
    """

    def __init__(self, *args: object) -> None:
        super().__init__(*args)

    @staticmethod
    def enforce_literal(
        option_selected: str,
        options: typing.Union[typing.List[str], typing.List[int]],
        param: str,
    ):

        if option_selected not in options:

            raise InvalidOptionError(
                param=param, value=option_selected, options=options
            )


class InvalidNumberOfArguments(MapillaryException):
    """
    Raised when an inappropriate number of parameters are passed to a function
    """

    def __init__(
        self,
        number_of_params_passed: int,
        actual_allowed_params: int,
        param: str,
        *args: object,
    ) -> None:
        super().__init__(*args)

        self.number_of_params_passed = number_of_params_passed
        self.actual_allowed_params = actual_allowed_params
        self.param = param

    def __str__(self):
        return (
            f'InvalidNumberOfArguments: The parameter, "{self.param}" was '
            f'passed "{self.number_of_params_passed}" items, when the max number of'
            f'allowed items are "{self.actual_allowed_params}"'
        )

    def __repr__(self):
        return (
            f"InvalidNumberOfArguments(number_of_params_passed={self.number_of_params_passed},"
            f"actual_allowed_params={self.actual_allowed_params}, param={self.param})"
        )



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.auth
~~~~~~~~~~~~~~~~~~~~~

This module contains the authorization logic for the client class of Mapillary, responsible
for keeping track of the session token set

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from functools import wraps

# Local imports
from zensvi.download.mapillary.models.client import Client
from zensvi.download.mapillary.models.exceptions import AuthError


def set_token(token: str) -> dict:
    """
    Allows the user to set access token to be able to interact with API v4

    :param token: Access token
    :return: Dictionary containing the access token
    """
    if len(token) < 1:
        raise AuthError("Token cannot be empty")
    try:
        Client.set_token(token)
    except AuthError:
        raise AuthError("Token is invalid")
    return {"token": "SUCCESS"}


def auth():
    """Wrap interface functions with logic for Client"""

    def auth_decorator(f):
        """Decorator function"""

        @wraps(f)
        def wrapper(*args, **kwargs):
            """
            :param args: Function arguments
            :param kwargs: Key word arguments
            :return: Return the specified function with args, kwargs
            """

            if Client.get_token() == "":
                # If empty, raise exception
                raise AuthError("Function called without setting the access token")

            # Return function called with arguments
            return f(*args, **kwargs)

        # Return wrapper
        return wrapper

    # Return decorator
    return auth_decorator



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.controllers.save
~~~~~~~~~~~~~~~~~~~~~~~~~~

This module implements the saving business logic functionalities of the Mapillary Python SDK.

For more information, please check out https://www.mapillary.com/developer/api-documentation/

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package Imports
import os
import json
import csv
from shapely.geometry import shape

# Local Imports
from zensvi.download.mapillary.utils.format import flatten_geojson
from zensvi.download.mapillary.utils.time import date_to_unix_timestamp
from zensvi.download.mapillary.utils.verify import check_file_name_validity


def save_as_csv_controller(data: str, path: str, file_name: str) -> None:
    """
    Save data as CSV to given file path

    :param data: The data to save as CSV
    :type data: str

    :param path: The path to save to
    :type path: str

    :param file_name: The file name to save as
    :type file_name: str

    :return: None
    :rtype: None
    """

    # Ensure that the geojson is a dictionary
    if isinstance(data, str):
        data = json.loads(data)

    # Flatten the geojson
    features = flatten_geojson(data)

    # Ensure that the file name is valid
    # Set the file name according to the given value. Default is
    # "mapillary_CURRENT_UNIX_TIMESTAMP_.csv"
    file_name = (
        file_name
        if (file_name is not None and check_file_name_validity(file_name))
        else f"mapillary_{date_to_unix_timestamp('*')}_"
    ) + ".csv"

    try:
        # Context manager for writing to file
        with open(os.path.join(path, file_name), "w", newline="") as file_path:
            # Enforce the header for field_names
            field_names = (
                ["ID", "WKT"]
                + [
                    key
                    for key in features[0].keys()
                    if key != "id" and key != "geometry"
                ]
                # organization_id may or may not exist in the flattened features
                # If it does exist, then its value will be reflected
                # If it does not exist, then it will be set to "NULL"
                + ["organization_id"]
            )

            # Create the csv writer
            writer = csv.DictWriter(file_path, fieldnames=field_names)

            # Write the header
            writer.writeheader()

            for feature in features:
                writer.writerow(
                    {
                        # ID is the id of the feature. It will always exist
                        # and will always be the first column
                        "ID": feature["id"],
                        # WKT is the geometry of the feature in well-known text format.
                        # It will always exist
                        "WKT": shape(feature["geometry"]).wkt,
                        # The rest of the columns are the features attributes
                        **{
                            key: feature[key]
                            for key in feature.keys()
                            if key != "id" and key != "geometry"
                        },
                        # organization_id is the id of the organization that the feature belongs to.
                        # If it does exist, then its value will be reflected
                        # If it does not exist, then it will be set to "NULL"
                        "organization_id": feature["organization_id"]
                        if "organization_id" in feature
                        else "NULL",
                    }
                )
    except Exception as e:
        # If there is an error, log it
        print(f"An error occurred: {e}")
    return None


def save_as_geojson_controller(data: str, path: str, file_name: str) -> None:
    """
    Save data as GeoJSON to given file path

    :param data: The data to save as GeoJSON
    :type data: str

    :param path: The path to save to
    :type path: str

    :param file_name: The file name to save as
    :type file_name: str

    :return: None
    :rtype: None
    """

    # Ensure that the geojson is a dictionary
    if isinstance(data, str):
        data = json.loads(data)

    # Ensure that the file name is valid
    # Set the file name according to the given value. Default is
    # "mapillary_CURRENT_UNIX_TIMESTAMP_.csv"
    file_name = (
        file_name
        if (file_name is not None and check_file_name_validity(file_name))
        else f"mapillary_{date_to_unix_timestamp('*')}_"
    ) + ".geojson"

    try:
        # Context manager for writing to file
        with open(os.path.join(path, file_name), "w") as file_path:
            json.dump(data, file_path, indent=4)
    except Exception as e:
        # If there is an error, log it
        print(e)

    return None



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.api.vector_tiles
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the class implementation of the VectorTile functionalities for the Vector Tile
aspect of the API v4 of Mapillary.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""


class VectorTiles:
    """
    Vector tiles provide an easy way to visualize vast amounts of data. Mapillary APIs are heavily
    based on vector tiles to provide the developers with flexibility to programmatically interact
    with the data they contain in custom ways. Vector tiles support filtering and querying rendered
    features. Mapillary vector tiles follow the Mapbox Tile Specification,
    https://docs.mapbox.com/vector-tiles/specification/
    """

    @staticmethod
    def get_overview_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original
        geometries (not computed) for the layer 'overview'

        This layer offers,

        1. zoom: 0 - 5 (inclusive)
        2. geometry: Point
        3. data source: images

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. id, int, ID of the image
        3. sequence_id, string, ID of the sequence this image belongs to
        4. is_pano, bool, if it is a panoramic image
        """

        return f"https://tiles.mapillary.com/maps/vtp/mly1_public/2/{z}/{x}/{y}/"

    @staticmethod
    def get_sequence_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original geometries (not computed) for the
        layer 'sequence'

        This layer offers,

        1. zoom: 6 - 14 (inclusive)
        2. geometry: LineString
        3. data source: images captured in a single collection, sorted by captured_at

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. id, string, ID  of the sequence (the legacy sequence key)
        3. image_id, int, ID of the 'best' (first) image representing the sequence
        4. organization_id, int, ID of the organization this image belongs to. It can be absent
        5. is_pano, bool, if it is a panoramic sequence
        """

        return f"https://tiles.mapillary.com/maps/vtp/mly1_public/2/{z}/{x}/{y}/"

    @staticmethod
    def get_image_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original geometries (not computed) for the
        layer 'image'

        This layer offers,

        1. zoom: 14
        2. geometry: Point
        3. data source: images

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. compass_angle, int, the compass angle of the image
        3. id, int, ID of the image
        4. sequence_id, string, ID of the sequence this image belongs to
        5. organization_id, int, ID of the organization this image belongs to. It can be absent
        6. is_pano, bool, if it is a panoramic image
        """

        return f"https://tiles.mapillary.com/maps/vtp/mly1_public/2/{z}/{x}/{y}/"

    @staticmethod
    def get_computed_overview_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original geometries
        (computed) for the layer 'overview'

        This layer offers,

        1. zoom: 0 - 5 (inclusive)
        2. geometry: Point
        3. data source: images

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. id, int, ID of the image
        3. sequence_id, string, ID of the sequence this image belongs to
        4. is_pano, bool, if it is a panoramic image
        """

        return (
            f"https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_computed_sequence_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original geometries (computed) for the
        layer 'sequence'

        This layer offers,

        1. zoom: 6 - 14 (inclusive)
        2. geometry: LineString
        3. data source: images captured in a single collection, sorted by captured_at

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. id, string, ID  of the sequence (the legacy sequence key)
        3. image_id, int, ID of the 'best' (first) image representing the sequence
        4. organization_id, int, ID of the organization this image belongs to. It can be absent
        5. is_pano, bool, if it is a panoramic sequence
        """

        return (
            f"https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_computed_image_layer(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        Contain positions of images and sequences with original geometries (computed) for the
        layer 'image'

        This layer offers,

        1. zoom: 14
        2. geometry: Point
        3. data source: images

        With the following properties,

        1. captured_at, int, timestamp in ms since epoch
        2. compass_angle, int, the compass angle of the image
        3. id, int, ID of the image
        4. sequence_id, string, ID of the sequence this image belongs to
        5. organization_id, int, ID of the organization this image belongs to. It can be absent
        6. is_pano, bool, if it is a panoramic image
        """

        return (
            f"https://tiles.mapillary.com/maps/vtp/mly1_computed_public/2/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_map_feature_point(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        These tiles represent positions of map features which are detected on the Mapillary platform
        and are not traffic signs.

        This layer offers,

        1. zoom: 14
        2. geometry: Point
        3. data source: map features

        With the following resultant properties,

        1. id, int, ID of the image
        2. value, string, name of the class which this object represent
        3. first_seen_at, int, timestamp in ms since epoch, capture time of the earliest image on
            which the detection contribute to this map feature
        4. last_seen_at, int, timestamp in ms since epoch, capture time of the latest image on which
            the detection contribute to this map feature
        """

        return (
            f"https://tiles.mapillary.com/maps/vtp/mly_map_feature_point/2/{z}/{x}/{y}/"
        )

    @staticmethod
    def get_map_feature_traffic_sign(
        x: float,
        y: float,
        z: float,
    ) -> str:
        """
        These tiles represent positions of map features which are detected on the Mapillary
        platform and are traffic signs.

        The tile metadata is exactly the same as Map feature tiles, points, except that the
        layer name is traffic_sign.

        This layer offers,

        1. zoom: 14
        2. geometry: Point
        3. data source: map features

        With the following properties,

        1. id, int, ID of the image
        2. value, string, name of the class which this object represent
        3. first_seen_at, int, timestamp in ms since epoch, capture time of the earliest image on
            which the detection contribute to this map feature
        4. last_seen_at, int, timestamp in ms since epoch, capture time of the latest image on
            which the detection contribute to this map feature
        """

        return f"https://tiles.mapillary.com/maps/vtp/mly_map_feature_traffic_sign/2/{z}/{x}/{y}/"



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.utils.format
======================

This module deals with converting data to and from different file formats.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Package imports
import base64
import json
import typing
import mapbox_vector_tile
from collections.abc import MutableMapping
from typing import Union

# Local imports
# # Models
from zensvi.download.mapillary.models.geojson import Coordinates, GeoJSON


def feature_to_geojson(json_data: dict) -> dict:
    """
    Converts feature into a GeoJSON, returns output

    From::

        >>> {'geometry': {'type': 'Point', 'coordinates': [30.003755665554, 30.985948744314]},
        ... 'id':'506566177256016'}

    To::

        >>> {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry': {'type':
        ... 'Point','coordinates': [30.98594605922699, 30.003757307208872]}, 'properties': {}}]}

    :param json_data: The feature as a JSON
    :type json_data: dict

    :return: The formatted GeoJSON
    :rtype: dict
    """

    # The geometry property will always be present
    keys = [key for key in json_data.keys() if key != "geometry"]

    feature = {"type": "Feature", "geometry": {}, "properties": {}}
    # Make sure that all keys exist and retrieve their values if specified

    for key in keys:
        feature["properties"][key] = json_data[key]

    return feature


def join_geojson_with_keys(
    geojson_src: dict,
    geojson_src_key: str,
    geojson_dest: dict,
    geojson_dest_key: str,
) -> dict:
    """
    Combines two GeoJSONS based on the similarity of their specified keys, similar to
    the SQL join functionality

    :param geojson_src: The starting GeoJSO source
    :type geojson_src: dict

    :param geojson_src_key: The key in properties specified for the GeoJSON source
    :type geojson_src_key: str

    :param geojson_dest: The GeoJSON to merge into
    :type geojson_dest: dict

    :param geojson_dest_key: The key in properties specified for the GeoJSON to merge into
    :type geojson_dest_key: dict

    :return: The merged GeoJSON
    :rtype: dict

    Usage::

        >>> join_geojson_with_keys(
        ...     geojson_src=geojson_src,
        ...     geojson_src_key='id',
        ...     geojson_dest=geojson_dest,
        ...     geojson_dest_key='id'
        ... )
    """

    # Go through the feature set in the src geojson
    for from_features in geojson_src["features"]:

        # Go through the feature set in the dest geojson
        for into_features in geojson_dest["features"]:

            # If either of the geojson features do not contain
            # their respective assumed keys, continue
            if (
                geojson_dest_key not in into_features["properties"]
                or geojson_src_key not in from_features["properties"]
            ):
                continue

            # Checking if two IDs match up
            if int(from_features["properties"][geojson_src_key]) == int(
                into_features["properties"][geojson_dest_key]
            ):

                # Firstly, extract the properties that exist in the
                # src_geojson for that feature
                old_properties = [key for key in from_features["properties"].keys()]

                # Secondly, extract the properties that exist in the
                # dest_json for that feature
                new_properties = [key for key in into_features["properties"].keys()]

                # Going through the old properties in the features of src_geojson
                for new_property in new_properties:

                    # Going through the new properties in the features of dest_geojson
                    if new_property not in old_properties:
                        # Put the new_feature
                        from_features["properties"][new_property] = old_properties[
                            "properties"
                        ][new_property]

    return geojson_src


def geojson_to_features_list(json_data: dict) -> list:
    """
    Converts a decoded output GeoJSON to a list of feature objects

    The purpose of this formatting utility is to obtain a list of individual features for
    decoded tiles that can be later extended to the output GeoJSON

    From::

        >>> {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry':
        ... {'type': 'Point','coordinates': [30.98594605922699, 30.003757307208872]},
        ... 'properties': {}}]}

    To::

        >>> [{'type': 'Feature', 'geometry': {'type': 'Point',
        ... 'coordinates': [30.98594605922699, 30.003757307208872]}, 'properties': {}}]

    :param json_data: The given json data
    :type json_data: dict

    :return: The feature list
    :rtype: list
    """

    return json_data["features"]


def merged_features_list_to_geojson(features_list: list) -> str:
    """
    Converts a processed features list (i.e. a features list with all the needed features merged
    from multiple tiles) into a fully-featured GeoJSON

    From::

        >>> [{'type': 'Feature', 'geometry': {'type': 'Point',
        ... 'coordinates': [30.98594605922699, 30.003757307208872]}, 'properties': {}}, ...]

    To::

        >>> {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry':
        ... {'type': 'Point','coordinates': [30.98594605922699, 30.003757307208872]},
        ... 'properties': {}}, ...]}

    :param features_list: a list of processed features merged from different tiles within a bbox
    :type features_list: list

    :return: GeoJSON string formatted with all the extra commas removed.
    :rtype: str
    """

    return json.dumps({"type": "FeatureCollection", "features": features_list})


def detection_features_to_geojson(feature_list: list) -> dict:
    """
    Converts a preprocessed list (i.e, features from the detections of either images or
    map_features from multiple segments) into a fully featured GeoJSON

    :param feature_list: A list of processed features merged from different segments within a
        detection
    :type feature_list: list

    :return: GeoJSON formatted as expected in a detection format
    :rtype: dict

    Example::

        >>> # From
        >>> [{'created_at': '2021-05-20T17:49:01+0000', 'geometry':
        ... 'GjUKBm1weS1vchIVEgIAABgDIg0JhiekKBoqAABKKQAPGgR0eXBlIgkKB3BvbHlnb24ogCB4AQ==',
        ... 'image': {'geometry': {'type': 'Point', 'coordinates': [-97.743279722222,
        ... 30.270651388889]}, 'id': '1933525276802129'}, 'value': 'regulatory--no-parking--g2',
        ... 'id': '1942105415944115'}, ... ]
        >>> # To
        >>> {'type': 'FeatureCollection', 'features': [{'type': 'Feature', 'geometry':
        ... {'type': 'Point', 'coordinates': [-97.743279722222, 30.270651388889]}, 'properties': {
        ... 'image_id': '1933525276802129', 'created_at': '2021-05-20T17:49:01+0000',
        ... 'pixel_geometry':
        ... 'GjUKBm1weS1vchIVEgIAABgDIg0JhiekKBoqAABKKQAPGgR0eXBlIgkKB3BvbHlnb24ogCB4AQ==',
        ... 'value': 'regulatory--no-parking--g2', 'id': '1942105415944115' } }, ... ]}
    """

    resulting_geojson = {
        # FeatureCollection type
        "type": "FeatureCollection",
        # List of features
        "features": [
            # Feature generation from feature_list
            {
                # Type is 'Feature'
                "type": "Feature",
                # Let 'geometry' be the `image` key, defaults to {} is `image` not in feature
                "geometry": {
                    "type": "Point",
                    "coordinates": feature["image"]["geometry"]["coordinates"],
                }
                if "image" in feature
                else {},
                # Property list
                "possible_none_properties": {
                    # Only if "image" was specified in the `fields` of the endpoint, else None
                    "image_id": feature["image"]["id"]
                    if "image" in "feature"
                    else None,
                    # Only if "created_at" was specified in the `fields` of the endpoint, else None
                    "created_at": feature["created_at"]
                    if "created_at" in feature
                    else None,
                    # Only if "geometry" was specified in the `fields` of the endpoint, else None
                    "pixel_geometry": feature["geometry"]
                    if "geometry" in feature
                    else None,
                    # Only if "value" was specified in the `fields` of the endpoint, else None
                    "value": feature["value"] if "value" in feature else None,
                    # "id" is always available in the response
                    "id": feature["id"],
                },
                "properties": {},
            }
            # Going through the given of features
            for feature in feature_list
        ],
    }

    # The next logic below removes features that defaulted to None

    # Through each feature in the resulting features
    for _feature in resulting_geojson["features"]:

        # Going through each property in the feature
        for key, value in _feature["possible_none_properties"].items():

            # If the _property is not None
            if value is not None:

                # Add that property to the _feature
                _feature["properties"][key] = value

    for _feature in resulting_geojson["features"]:

        del _feature["possible_none_properties"]

    # Finally return the output
    return resulting_geojson


def flatten_geojson(geojson: dict) -> list:
    """
    Flattens a GeoJSON dictionary to a dictionary with only the relevant keys.
    This is useful for writing to a CSV file.

    Output Structure::

        >>> {
        ...     "geometry": {
        ...         "type": "Point",
        ...         "coordinates": [71.45343, 12.523432]
        ...     },
        ...     "first_seen_at": "UNIX_TIMESTAMP",
        ...     "last_seen_at": "UNIX_TIMESTAMP",
        ...     "value": "regulatory--no-parking--g2",
        ...     "id": "FEATURE_ID",
        ...     "image_id": "IMAGE_ID"
        ... }

    :param geojson: The GeoJSON to flatten
    :type geojson: dict

    :return: A flattened GeoJSON
    :rtype: dict

    Note,
        1. The `geometry` key is always present in the output
        2. The properties are flattened to the following keys:
            - "first_seen_at"   (optional)
            - "last_seen_at"    (optional)
            - "value"           (optional)
            - "id"              (required)
            - "image_id"        (optional)
            - etc.
        3. If the 'geometry` type is `Point`, two more properties will be added:
            - "longitude"
            - "latitude"

    *TODO*: Further testing needed with different geometries, e.g., Polygon, etc.
    """

    for feature in geojson["features"]:
        # Check if the geometry is a Point
        if feature["geometry"]["type"] == "Point":
            # Add longitude and latitude properties to the feature
            feature["properties"]["longitude"] = feature["geometry"]["coordinates"][0]
            feature["properties"]["latitude"] = feature["geometry"]["coordinates"][1]

    # Return the flattened geojson
    return [
        {"geometry": _feature["geometry"], **_feature["properties"]}
        for _feature in geojson["features"]
    ]


def geojson_to_polygon(geojson: dict) -> GeoJSON:
    """
    Converts a GeoJSON into a collection of only geometry coordinates for the purpose of
    checking whether a given coordinate point exists within a shapely polygon

    From::

        >>> {
        ...     "type": "FeatureCollection",
        ...     "features": [
        ...         {
        ...             "geometry": {
        ...                 "coordinates": [
        ...                     -80.13069927692413,
        ...                     25.78523699486192
        ...                 ],
        ...                 "type": "Point"
        ...             },
        ...             "properties": {
        ...                 "first_seen_at": 1422984049000,
        ...                 "id": 481978503020355,
        ...                 "last_seen_at": 1422984049000,
        ...                 "value": "object--street-light"
        ...             },
        ...             "type": "Feature"
        ...         },
        ...         {
        ...             "geometry": {
        ...                 "coordinates": [
        ...                     -80.13210475444794,
        ...                     25.78362849816017
        ...                 ],
        ...                 "type": "Point"
        ...             },
        ...             "properties": {
        ...                 "first_seen_at": 1423228306666,
        ...                 "id": 252538103315239,
        ...                 "last_seen_at": 1423228306666,
        ...                 "value": "object--street-light"
        ...             },
        ...             "type": "Feature"
        ...         },
        ...         ...
        ...     ]
        ... }

    To::

        >>> {
        ... "type": "FeatureCollection",
        ... "features": [
        ...         {
        ...             "type": "Feature",
        ...             "properties": {},
        ...             "geometry": {
        ...                 "type": "Polygon",
        ...                 "coordinates": [
        ...                     [
        ...                         [
        ...                             7.2564697265625,
        ...                             43.69716905314008
        ...                         ],
        ...                         [
        ...                             7.27020263671875,
        ...                             43.69419030566581
        ...                         ],
        ...                         ...
        ...                     ]
        ...                 ]
        ...             }
        ...         }
        ...     ]
        ... }

    :param geojson: The input GeoJSON
    :type geojson: dict

    :return: A geojson of the format mentioned under 'To'
    :rtype: dict
    """

    return GeoJSON(
        geojson={
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "properties": {},
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                # Double listed on purpose. See above example under 'To'
                                feature["geometry"]["coordinates"]
                                for feature in geojson["features"]
                            ]
                        ],
                    },
                }
            ],
        }
    )


def flatten_dictionary(
    data: typing.Union[dict, MutableMapping],
    parent_key: str = "",
    sep: str = "_",
) -> dict:
    """
    Flattens dictionaries

    From::

        >>> {'mpy-or': {'extent': 4096, 'version': 2, 'features': [{'geometry': {'type':
        ... 'Polygon', 'coordinates': [[[2402, 2776], [2408, 2776]]]}, 'properties': {}, 'id': 1,
        ... 'type': 3}]}}

    To::

        >>> {'mpy-or_extent': 4096, 'mpy-or_version': 2, 'mpy-or_features': [{'geometry':
        ... {'type':'Polygon', 'coordinates': [[[2402, 2776], [2408, 2776]]]}, 'properties':
        ... {}, 'id': 1,'type': 3}]}

    :param data: The dictionary itself
    :type data: dict

    :param parent_key: The root key to start from
    :type parent_key: str

    :param sep: The separator
    :type sep: str

    :return: A flattened dictionary
    :rtype: dict
    """

    # Final results
    items = []

    # Traversing dictionary items
    for key, value in data.items():

        # Getting the new key
        new_key = parent_key + sep + key if parent_key else key

        # Checking value instance
        # MutableMapping makes this Python 2.6+ compatible
        if isinstance(value, MutableMapping):

            # Extending items list
            items.extend(flatten_dictionary(value, new_key, sep=sep).items())

        # If not instance
        else:

            # Append to the list
            items.append((new_key, value))

    # Converting to dictionary, and returning
    return dict(items)


def normalize_list(coordinates: list, width: int = 4096, height: int = 4096) -> list:
    """
    Normalizes a list of coordinates with the respective width and the height

    From::

        >>> [[[2402, 2776], [2408, 2776]]]

    To::

        >>> normalize_list(coordinates)
        ... # [[[0.58642578125, 0.677734375], [0.587890625, 0.677734375]]]

    :param coordinates: The coordinate list to normalize
    :type coordinates: list

    :param width: The width of the coordinates to normalize with, defaults to 4096
    :type width: int

    :param height: The height of the coordinates to normalize with, defaults to 4096
    :type height: int

    :return: The normalized list
    :rtype: list
    """

    # Extracting the list from the coordinates
    coordinates = coordinates[0]

    # Initializing the coordinate list
    new_coordinates = []

    # Going through each coordinate pair
    for coordinate_pair in coordinates:

        # If it is already normalized ...
        if 0 <= coordinate_pair[0] <= 1 and 0 <= coordinate_pair[1] <= 1:
            # ... then append as is
            new_coordinates.append(coordinate_pair)

        # Appending the coordinates
        new_coordinates.append(
            # Appending a list pair of the width, height
            [coordinate_pair[0] / width, coordinate_pair[1] / height]
        )

    # Returning the results
    return [new_coordinates]


def decode_pixel_geometry(
    base64_string: str, normalized: bool = True, width: int = 4096, height: int = 4096
) -> dict:
    """
    Decodes the pixel geometry, and return the coordinates, which can be specified to be
    normalized

    :param base64_string: The pixel geometry encoded as a vector tile
    :type base64_string: str

    :param normalized: If normalization is required, defaults to True
    :type normalized: bool

    :param width: The width of the pixel geometry, defaults to 4096
    :type width: int

    :param height: The height of the pixel geometry, defaults to 4096
    :type height: int

    :return: A dictionary with coordinates as key, and value as the normalized list
    :rtype: list
    """

    # The data retrieved after being decoded as base64
    data = base64.decodebytes(base64_string.encode("utf-8"))

    # Getting the results from mapbox_vector_tile
    result = mapbox_vector_tile.decode(data)

    # Flattening the resultant dictionary
    flattened = flatten_dictionary(result)

    # Init geometry var
    geometry = None

    # For key in the flattened keys
    for key in flattened.keys():

        # If the key contains the word "features"
        if "features" in key:
            # Get the geometry
            geometry = flattened[key][0]

    # Extract the coordinate list
    coordinate_list = geometry["geometry"]["coordinates"]

    # Return output
    return (
        # Return coordinates as normalized values, if normalized is true
        {"coordinates": normalize_list(coordinate_list, width=width, height=height)}
        if normalized
        # Else return un-normalized coordinates
        else {"coordinates": coordinate_list}
    )


def decode_pixel_geometry_in_geojson(
    geojson: typing.Union[dict, GeoJSON],
    normalized: bool = True,
    width: int = 4096,
    height: int = 4096,
) -> GeoJSON:
    """
    Decodes all the pixel_geometry

    :param geojson: The GeoJSON representation to be decoded

    :param normalized: If normalization is required, defaults to True
    :type normalized: bool

    :param width: The width of the pixel geometry, defaults to 4096
    :type width: int

    :param height: The height of the pixel geometry, defaults to 4096
    :type height: int
    """

    # If geojson is of type GeoJSON, convert to dict
    if isinstance(geojson, GeoJSON):
        geojson: dict = geojson.to_dict()

    # Remove pass by reference for idempotent principle
    data = geojson.copy()

    # Iterate through the features
    for feature in data["features"]:
        # For the pixel_geometry feature, decode it
        feature["properties"]["pixel_geometry"] = decode_pixel_geometry(
            # Get the base64_string
            base64_string=feature["properties"]["pixel_geometry"],
            # Normalization variable
            normalized=normalized,
            # Width param
            width=width,
            # Height param
            height=height,
        )

    # Return output as GeoJSON
    return GeoJSON(geojson=data)


def coord_or_list_to_dict(data: Union[Coordinates, list, dict]) -> dict:
    """
    Converts a Coordinates object or a coordinates list to a dictionary

    :param data: The coordinates to convert
    :type data: Union[Coordinates, list]

    :return: The dictionary representation of the coordinates
    :rtype: dict
    """

    if isinstance(data, dict):
        return data

    data_copy = None

    # If data is a list, convert to dict
    if isinstance(data, list):
        data_copy = {"lng": data[0], "lat": data[1]}

    # if data is a Coordinate object, convert to dict
    if isinstance(data, Coordinates):
        data_copy = data.to_dict()

    # Return the dictionary
    return data_copy


def polygon_feature_to_bbox_list(
    polygon: dict, is_bbox_list_required: bool = True
) -> typing.Union[list, dict]:
    """
    Converts a polygon to a bounding box

    The polygon below has been obtained from https://geojson.io/. If you have a polygon,
    with only 4 array elements, then simply take the first element and append it to the
    coordinates to obtain the below example.

    Usage::

        >>> from mapillary.utils.format import polygon_feature_to_bbox_list
        >>> bbox = polygon_feature_to_bbox_list(polygon={
        ...     "type": "Feature",
        ...     "properties": {},
        ...     "geometry": {
        ...         "type": "Polygon",
        ...         "coordinates": [
        ...             [
        ...                 [
        ...                   48.1640625,
        ...                   38.41055825094609
        ...                 ],
        ...                 [
        ...                   62.22656249999999,
        ...                   38.41055825094609
        ...                 ],
        ...                 [
        ...                   62.22656249999999,
        ...                   45.336701909968134
        ...                 ],
        ...                 [
        ...                   48.1640625,
        ...                   45.336701909968134
        ...                 ],
        ...                 [
        ...                   48.1640625,
        ...                   38.41055825094609
        ...                 ]
        ...             ]
        ...        ]
        ... })
        >>> bbox
        ... [62.22656249999999, 48.1640625, 38.41055825094609, 45.336701909968134]

    :param polygon: The polygon to convert
    :type polygon: dict

    :param is_bbox_list_required: Flag if bbox is required as a list. If true, returns a list,
    else returns a dict
    :type is_bbox_list_required: bool
    :default is_bbox_list_required: True

    :return: The bounding box
    :rtype: typing.Union[list, dict]
    """

    west = polygon["geometry"]["coordinates"][0][1][0]
    south = polygon["geometry"]["coordinates"][0][0][0]
    east = polygon["geometry"]["coordinates"][0][0][1]
    north = polygon["geometry"]["coordinates"][0][2][1]

    if is_bbox_list_required:
        return [west, south, east, north]

    return {"west", west, "south", south, "east", east, "north", north}


def bbox_to_polygon(bbox: typing.Union[list, dict]) -> dict:
    """
    Converts a bounding box dictionary to a polygon

    Usage::

        >>> from mapillary.utils.format import bbox_to_polygon
        >>> bbox = [62.22656249999999, 48.1640625, 38.41055825094609, 45.336701909968134]
        >>> polygon = bbox_to_polygon(bbox=bbox)
        >>> polygon
        ... {
        ...     "type": "Feature",
        ...     "properties": {},
        ...     "geometry": {
        ...         "type": "Polygon",
        ...         "coordinates": [
        ...             [
        ...                 [
        ...                   48.1640625,
        ...                   38.41055825094609
        ...                 ],
        ...                 [
        ...                   62.22656249999999,
        ...                   38.41055825094609
        ...                 ],
        ...                 [
        ...                   62.22656249999999,
        ...                   45.336701909968134
        ...                 ],
        ...                 [
        ...                   48.1640625,
        ...                   45.336701909968134
        ...                 ],
        ...                 [
        ...                   48.1640625,
        ...                   38.41055825094609
        ...                 ]
        ...             ]
        ...        ]
        ... })

    :param bbox: The bounding box to convert
    :type bbox: dict

    :return: The polygon
    :rtype: dict
    """

    # Initializing the polygon
    polygon = {
        "type": "Feature",
        "properties": {},
        "geometry": {
            "type": "Polygon",
            "coordinates": [[[0, 0], [0, 0], [0, 0], [0, 0], [0, 0]]],
        },
    }

    # If bbox is a list, convert to dict
    if isinstance(bbox, list):
        bbox = {"west": bbox[0], "south": bbox[1], "east": bbox[2], "north": bbox[3]}

    # For the top left to the bottom left
    polygon["geometry"]["coordinates"][0][0] = [bbox["south"], bbox["east"]]

    # from the bottom left to the bottom right
    polygon["geometry"]["coordinates"][0][1] = [bbox["west"], bbox["east"]]

    # from the bottom right to the top right
    polygon["geometry"]["coordinates"][0][2] = [bbox["west"], bbox["north"]]

    # from the top right to the top left
    polygon["geometry"]["coordinates"][0][3] = [bbox["south"], bbox["north"]]

    # from the top left to the top left
    polygon["geometry"]["coordinates"][0][4] = [bbox["south"], bbox["east"]]

    # Returning the results
    return polygon



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.api.__init__

This package contains all the API v4 endpoints provided with the Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import vector_tiles  # noqa: F401
from . import entities  # noqa: F401
from . import general  # noqa: F401



# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.api.entities
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This module contains the class implementation of the Entities API endpoints
as string, for the entity API endpoint aspect of the API v4 of Mapillary.

For more information, please check out https://www.mapillary.com/developer/api-documentation/.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

# Local imports
import datetime
import logging
import typing

# # Utils
from zensvi.download.mapillary.utils.time import is_iso8601_datetime_format

# # Models
# # # Exception Handling
from zensvi.download.mapillary.models.exceptions import InvalidFieldError, InvalidNumberOfArguments
from zensvi.download.mapillary.models.config import Config
from zensvi.download.mapillary.models.logger import Logger

logger: logging.Logger = Logger.setup_logger(name="mapillary.config.api.entities")


class Entities:
    """
    Each API call requires specifying the fields of the Entity you're interested in explicitly.
    A sample image by ID request which returns the id and a computed geometry could look as
    below. For each entity available fields are listed in the relevant sections. All IDs are
    unique and the underlying metadata for each entity is accessible at
    https://graph.mapillary.com/:id?fields=A,B,C. The responses are uniform and always return
    a single object, unless otherwise stated (collection endpoints). All collection endpoint
    metadata are wrapped in a {"data": [ {...}, ...]} JSON object.

    Usage::

        $ GET 'https://graph.mapillary.com/$IMAGE_ID?access_token=TOKEN&fields=id,computed_geometry'
        ... {
        ...     "id": "$IMAGE_ID",
        ...     "computed_geometry": {
        ...         "type": "Point",
        ...         "coordinates": [0, 0]
        ...     }
        ... }
    """

    @staticmethod
    def get_image(image_id: str, fields: list) -> str:
        """
        Represents the metadata of the image on the Mapillary platform with
        the following properties.

        Usage::

            >>> 'https://graph.mapillary.com/:image_id' # endpoint

        Fields::

            1. altitude - float, original altitude from Exif
            2. atomic_scale - float, scale of the SfM reconstruction around the image
            3. camera_parameters - array of float, intrinsic camera parameters
            4. camera_type - enum, type of camera projection (perspective, fisheye, or spherical)
            5. captured_at - timestamp, capture time
            6. compass_angle - float, original compass angle of the image
            7. computed_altitude - float, altitude after running image processing
            8. computed_compass_angle - float, compass angle after running image processing
            9. computed_geometry - GeoJSON Point, location after running image processing
            10. computed_rotation - enum, corrected orientation of the image
            11. exif_orientation - enum, orientation of the camera as given by the exif tag
                (see: https://sylvana.net/jpegcrop/exif_orientation.html)
            12. geometry - GeoJSON Point geometry
            13. height - int, height of the original image uploaded
            14. thumb_256_url - string, URL to the 256px wide thumbnail
            15. thumb_1024_url - string, URL to the 1024px wide thumbnail
            16. thumb_2048_url - string, URL to the 2048px wide thumbnail
            17. merge_cc - int, id of the connected component of images that were aligned together
            18. mesh - { id: string, url: string } - URL to the mesh
            19. quality_score - float, how good the image is (experimental)
            20. sequence - string, ID of the sequence
            21. sfm_cluster - { id: string, url: string } - URL to the point cloud
            22. width - int, width of the original image uploaded
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_image_fields(),
            endpoint="https://graph.mapillary.com/:image_id?fields=",
        )

        return f"https://graph.mapillary.com/{image_id}/?fields={','.join(fields)}"

    @staticmethod
    def get_images(
        image_ids: typing.Union[typing.List[str], typing.List[int]], fields: list
    ) -> str:
        """
        Represents the metadata of the image on the Mapillary platform with
        the following properties.

        Usage::

            >>> 'https://graph.mapillary.com/ids=ID1,ID2,ID3' # endpoint

        Parameters::

            A list of entity IDs separated by comma. The provided IDs must be in the same type
            (e.g. all image IDs, or all detection IDs)

        Fields::

            1. altitude - float, original altitude from Exif
            2. atomic_scale - float, scale of the SfM reconstruction around the image
            3. camera_parameters - array of float, intrinsic camera parameters
            4. camera_type - enum, type of camera projection (perspective, fisheye, or spherical)
            5. captured_at - timestamp, capture time
            6. compass_angle - float, original compass angle of the image
            7. computed_altitude - float, altitude after running image processing
            8. computed_compass_angle - float, compass angle after running image processing
            9. computed_geometry - GeoJSON Point, location after running image processing
            10. computed_rotation - enum, corrected orientation of the image
            11. exif_orientation - enum, orientation of the camera as given by the exif tag
                (see: https://sylvana.net/jpegcrop/exif_orientation.html)
            12. geometry - GeoJSON Point geometry
            13. height - int, height of the original image uploaded
            14. thumb_256_url - string, URL to the 256px wide thumbnail
            15. thumb_1024_url - string, URL to the 1024px wide thumbnail
            16. thumb_2048_url - string, URL to the 2048px wide thumbnail
            17. merge_cc - int, id of the connected component of images that were aligned together
            18. mesh - { id: string, url: string } - URL to the mesh
            19. quality_score - float, how good the image is (experimental)
            20. sequence - string, ID of the sequence
            21. sfm_cluster - { id: string, url: string } - URL to the point cloud
            22. width - int, width of the original image uploaded

        Raises::

            InvalidNumberOfArguments - if the number of ids passed is 0 or greater than 50
        """

        # TODO: while this function should not be responsible to check if
        # all of the IDs passed are image_ids or detections_ids, this comment
        # should be to remind that this logic should belong elsewhere or to integrated
        # in the future

        if len(image_ids) == 0 or len(image_ids) > 50:
            raise InvalidNumberOfArguments(
                number_of_params_passed=len(image_ids),
                actual_allowed_params=50,
                param="image_ids",
            )

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_image_fields(),
            endpoint="https://graph.mapillary.com/ids=",
        )

        return f"https://graph.mapillary.com/ids={','.join(image_ids)}?fields={','.join(fields)}"

    @staticmethod
    def search_for_images(  # noqa: C901, 'search_for_images is too complex'
        bbox: typing.List[float],
        start_captured_at: typing.Optional[datetime.datetime] = None,
        end_captured_at: typing.Optional[datetime.datetime] = None,
        limit: typing.Optional[int] = None,
        organization_id: typing.Union[
            typing.Optional[int], typing.Optional[str]
        ] = None,
        sequence_id: typing.Optional[typing.List[int]] = None,
        fields: typing.Optional[list] = [],
    ) -> str:
        """
        Represents the metadata of the image on the Mapillary platform with
        the following properties.

        Output Format::

            >>> 'https://graph.mapillary.com/search?bbox=LONG1,LAT1,LONG2,LAT2' # endpoint
            >>> 'https://graph.mapillary.com/search?bbox=LONG1,LAT1,LONG2,LAT2&start_time='
            'START_TIME' # endpoint
            >>> 'https://graph.mapillary.com/search?bbox=LONG1,LAT1,LONG2,LAT2&start_time='
            'START_TIME&end_time=END_TIME' # endpoint
            >>> 'https://graph.mapillary.com/search?bbox=LONG1,LAT1,LONG2,LAT2&start_time='
            'START_TIME&end_time=END_TIME&limit=LIMIT' # endpoint
            >>> 'https://graph.mapillary.com/search/images?bbox=LONG1,LAT1,LONG2,LAT2&start_time'
            '=START_TIME&end_time=END_TIME&limit=LIMIT&organization_id=ORGANIZATION_ID&'
            'sequence_id=SEQUENCE_ID1' # endpoint
            >>> 'https://graph.mapillary.com/search/images?bbox=LONG1,LAT1,LONG2,LAT2&start_time='
            'START_TIME&end_time=END_TIME&limit=LIMIT&organization_id=ORGANIZATION_ID&sequence_id'
            '=SEQUENCE_ID1,SEQUENCE_ID2,SEQUENCE_ID3' # endpoint

        Usage::

            >>> from mapillary.config.api.entities import Entities
            >>> bbox = [-180, -90, 180, 90]
            >>> start_captured_at = datetime.datetime(2020, 1, 1, 0, 0, 0)
            >>> end_captured_at = datetime.datetime(2022, 1, 1, 0, 0, 0)
            >>> organization_id = 123456789
            >>> sequence_ids = [123456789, 987654321]
            >>> Entities.search_for_images(bbox=bbox) # endpoint
            'https://graph.mapillary.com/search?bbox=-180,-90,180,90' # endpoint
            >>> Entities.search_for_images(bbox=bbox, start_captured_at=start_captured_at)
            'https://graph.mapillary.com/search?bbox=-180,-90,180,90&start_time=' # endpoint
            >>> Entities.search_for_images(bbox=bbox,
            ... start_captured_at=start_captured_at, end_captured_at=end_captured_at)
            'https://graph.mapillary.com/search?bbox=-180,-90,180,90&start_time=&'
            'end_time=' # endpoint
            >>> Entities.search_for_images(bbox=bbox,
            ... start_captured_at=start_captured_at, end_captured_at=end_captured_at,
            ... limit=100)
            'https://graph.mapillary.com/search?bbox=-180,-90,180,90&start_time=&end_time=&limit'
            '=100' # endpoint
            >>> Entities.search_for_images(bbox=bbox,
            ... start_captured_at=start_captured_at, end_captured_at=end_captured_at,
            ... limit=100, organization_id=organization_id, sequence_id=sequence_ids)
            'https://graph.mapillary.com/search/images?bbox=-180,-90,180,90&start_time=&end_time'
            '=&limit=100&organization_id=1234567890&sequence_id=1234567890' # endpoint

        :param bbox: float,float,float,float: filter images in the bounding box. Specify in this
        order: left, bottom, right, top (or minLon, minLat, maxLon, maxLat).
        :type bbox: typing.Union[typing.List[float], typing.Tuple[float, float, float, float],
        list, tuple]

        :param start_captured_at: filter images captured after. Specify in the ISO 8601 format.
        For example: "2022-08-16T16:42:46Z".
        :type start_time: typing.Union[typing.Optional[datetime.datetime], typing.Optional[str]]
        :default start_captured_at: None

        :param end_captured_at: filter images captured before. Same format as
        "start_captured_at".
        :type end_time: typing.Union[typing.Optional[datetime.datetime], typing.Optional[str]]
        :default end_captured_at: None

        :param limit: limit the number of images returned. Max and default is 2000. The 'default'
        here means the default value of `limit` assumed on the server's end if the limit param
        is not passed. In other words, if the `limit` parameter is set to `None`, the server will
        assume the `limit` parameter to be 2000, which is the same as setting the `limit`
        parameter to 2000 explicitly.
        :type limit: typing.Optional[int]
        :default limit: None

        :param organization_id: filter images contributed to the specified organization Id.
        :type organization_id: typing.Optional[int]
        :default organization_id: None

        :param sequence_id: filter images in the specified sequence Ids (separated by commas),
        For example, "[1234567890,1234567891,1234567892]".
        :type sequence_id: typing.Optional[typing.List[int], int]
        :default sequence_id: None

        :param fields: filter the fields returned. For example, "['atomic_scale', 'altitude',
        'camera_parameters']". For more information, see
        https://www.mapillary.com/developer/api-documentation/#image. To get list of all possible
        fields, please use Entities.get_image_fields()
        :type fields: typing.Optional[typing.List[str]]
        :default fields: []

        :return: endpoint for searching an image
        :rtype: str
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_image_fields(),
            endpoint="https://graph.mapillary.com/images?bbox=,:parameters=,:fields=",
        )

        parameter_string: str = ""

        parameters = {
            "start_captured_at": start_captured_at,
            "end_captured_at": end_captured_at,
            "limit": limit,
            "organization_id": organization_id,
            "sequence_id": sequence_id,
        }

        # For each item in the given parameters ...
        for key, value in parameters.items():

            # ... if it is not None ...
            if value is not None:

                # ... if the key is about time ...
                if key in ["start_captured_at", "end_captured_at"]:

                    # ... if the datatype is a string ...
                    if isinstance(value, str):

                        # ... check if the string is a valid datetime in the ISO8601 format ...
                        if not is_iso8601_datetime_format(value):

                            # ... if not, raise an error ...
                            if Config.use_strict:

                                # Raising ValueError if strict mode is enabled
                                raise ValueError(
                                    f"""{key} must be in the ISO 8601 format. For example:
                                    '2022-08-16T16:42:46Z'."""
                                )

                            else:

                                logger.warning(
                                    f"{key} must be in the ISO 8601 format. For example:"
                                    f"'2022-08-16T16:42:46Z'. Disregarding {key} parameter."
                                    "Continuing without strict mode enabled."
                                )

                            # ... if not, just move on - no assumptions on the `date string` ...
                            continue

                    # ... if the value is a valid datetime object ...
                    elif isinstance(value, datetime.datetime):

                        # ... convert it to the ISO 8601 format required ...
                        value = value.strftime("%Y-%m-%dT%H:%M:%SZ")

                # ... if the key is limit ...
                if key == "limit":

                    # Check if it is within limits
                    if value > 2000:

                        # Log warning if not, waring mode is enabled - logger object declared
                        # globally is used
                        logger.warning(f"{key} is greater than 2000. Setting to 2000.")

                        # Set the value to 2000
                        value = 2000

                # ... if the key is sequence_id ...
                if key == "sequence_id":
                    # ... convert the list into string ...
                    value = ",".join(map(str, value))

                # ... add it to the parameter string
                parameter_string += f"&{key}={value}"

        # Return the endpoint
        return (
            f"https://graph.mapillary.com/images?bbox={','.join([str(val) for val in bbox])}"
            f"{parameter_string if parameter_string != '' else ''}"
            f"{'&fields=' + ','.join(fields) if fields != [] else ''}"
        )

    @staticmethod
    def get_image_fields() -> list:
        """
        Gets list of possible image fields

        :return: Image field list
        :rtype: list
        """

        return [
            "altitude",
            "atomic_scale",
            "camera_parameters",
            "camera_type",
            "captured_at",
            "compass_angle",
            "computed_altitude",
            "computed_compass_angle",
            "computed_geometry",
            "computed_rotation",
            "exif_orientation",
            "geometry",
            "height",
            "thumb_256_url",
            "thumb_1024_url",
            "thumb_2048_url",
            "merge_cc",
            "mesh",
            "quality_score",
            "sequence",
            "sfm_cluster",
            "width",
        ]

    @staticmethod
    def get_map_feature(map_feature_id: str, fields: list) -> str:
        """
        These are objects with a location which have been derived from
        multiple detections in multiple images.

        Usage::

            >>> 'https://graph.mapillary.com/:map_feature_id' # endpoint

        Fields::

            1. first_seen_at - timestamp, timestamp of the least recent detection
                contributing to this feature
            2. last_seen_at - timestamp, timestamp of the most recent detection
                contributing to this feature
            3. object_value - string, what kind of map feature it is
            4. object_type - string, either a traffic_sign or point
            5. geometry - GeoJSON Point geometry
            6. images - list of IDs, which images this map feature was derived from
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_map_feature_fields(),
            endpoint="https://graph.mapillary.com/:map_feature_id?fields=",
        )

        return (
            f"https://graph.mapillary.com/{map_feature_id}/?fields={','.join(fields)}"
        )

    @staticmethod
    def get_map_feature_fields() -> list:
        """
        Gets map feature fields

        :return: Possible map feature fields
        :rtype: list
        """

        return [
            "first_seen_at",
            "last_seen_at",
            "object_value",
            "object_type",
            "geometry",
            "images",
        ]

    @staticmethod
    def get_detection_with_image_id(
        image_id: str,
        fields: list,
    ) -> str:
        """
        Represent an object detected in a single image. For convenience
        this version of the API serves detections as collections. They can be
        requested as a collection on the resource (e.g. image) they contribute
        or belong to.

        Usage::

            >>> 'https://graph.mapillary.com/:image_id/detections'
            >>> # detections in the image with ID image_id

        Fields::

            1. created_at - timestamp, when was this detection created
            2. geometry - string, base64 encoded polygon
            3. image - object, image the detection belongs to
            4. value - string, what kind of object the detection represents
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_detection_with_image_id_fields(),
            endpoint="https://graph.mapillary.com/:image_id/detections/?fields=",
        )

        return f"https://graph.mapillary.com/{image_id}/detections/?fields={','.join(fields)}"

    @staticmethod
    def get_detection_with_image_id_fields() -> list:
        """
        Gets list of possible detections for image ids

        :return: Possible detection parameters
        :rtype: list
        """

        return ["created_at", "geometry", "image", "value"]

    @staticmethod
    def get_detection_with_map_feature_id(
        map_feature_id: str,
        fields: list,
    ) -> str:
        """
        Represent an object detected in a single image. For convenience
        this version of the API serves detections as collections. They can be
        requested as a collection on the resource (e.g. map feature) they
        contribute or belong to.

        Usage::

            >>> 'https://graph.mapillary.com/:map_feature_id/detections'
            >>> # detections in the image with ID map_feature_id

        Fields::

            1. created_at - timestamp, when was this detection created
            2. geometry - string, base64 encoded polygon
            3. image - object, image the detection belongs to
            4. value - string, what kind of object the detection represents
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_detection_with_map_feature_id_fields(),
            endpoint="https://graph.mapillary.com/:map_feature_id/detections/?fields=",
        )

        return f"https://graph.mapillary.com/{map_feature_id}/detections/?fields={','.join(fields)}"

    @staticmethod
    def get_detection_with_map_feature_id_fields() -> list:
        """
        Gets list of possible field parameters for map features

        :return: Map feature detection fields
        :rtype: list
        """

        return ["created_at", "geometry", "image", "value"]

    @staticmethod
    def get_organization_id(
        organization_id: str,
        fields: list,
    ) -> str:
        """
        Represents an organization which can own the imagery if users
        upload to it

        Usage::

            >>> 'https://graph.mapillary.com/:organization_id' # endpoint

        Fields::

            1. slug - short name, used in URLs
            2. name - nice name
            3. description - public description of the organization
        """

        fields = Entities.__field_validity(
            given_fields=fields,
            actual_fields=Entities.get_organization_id_fields(),
            endpoint="https://graph.mapillary.com/:organization_id?fields=",
        )

        return (
            f"https://graph.mapillary.com/{organization_id}/?fields={','.join(fields)}"
        )

    @staticmethod
    def get_organization_id_fields() -> list:
        """
        Gets list of possible organization id fields

        :return: Possible organization fields
        :rtype: list
        """

        return ["slug", "name", "description"]

    @staticmethod
    def get_sequence(
        sequence_id: str,
    ) -> str:
        """
        Represents a sequence of Image IDs ordered by capture time

        Usage::

            >>> 'https://graph.mapillary.com/image_ids?sequence_id=XXX'
            >>> # endpoint

        Fields::

            1. id - ID of the image belonging to the sequence
        """

        return f"https://graph.mapillary.com/image_ids?sequence_id={sequence_id}"

    @staticmethod
    def __field_validity(
        given_fields: list, actual_fields: list, endpoint: str
    ) -> list:
        """
        Checks if the given_fields are the actual correct fields for the given endpoint
        Compares against the list provided in `actual_fields`

        :param given_fields: The fields given as argument to check in
        :type given_fields: list

        :param actual_fields: The fields to check against
        :type actual_fields: list

        :param endpoint: The endpoint that is being targeted
        :type endpoint: str

        :raises InvalidFieldError: Raised when an API endpoint is passed invalid field elements

        :return: The given_fields if everything is correct
        :rtype: list
        """

        if len(given_fields) == 0:
            return given_fields  # empty list: []

        # Converting the given_fields into lowercase
        given_fields = [field.lower() for field in given_fields]

        # Enforce the existence of 'geometry' in the given_fields
        if "geometry" not in given_fields:
            given_fields.append("geometry")

        # Going through all the given fields
        for field in given_fields:

            # If 'all' is encountered ...
            if field == "all":

                # ... simply return the actual_fields list
                return actual_fields

            # If a field does not exist in the actual_fields ...
            if field not in actual_fields:

                # Raise an InvalidFieldError error
                raise InvalidFieldError(
                    # Specifying what endpoint was specified
                    endpoint=endpoint,
                    # Specify what field triggered the exception
                    field=field,
                )

        # If no error occurred, and all the fields are correct, return the given_fields
        return given_fields
